define('StreamXMLParser', ['exports'], function (exports) { 'use strict';

// import {arrayFrom, arrayConcat} from './collection';

function listToNode(list, adapter) {
	var node = adapter.initRoot();
	adapter.childrenSet(node, list);
	return node;
}

function treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree, targetTree) {
	if (!ctx) { ctx = {}; }
	if (!targetAdapter) { targetAdapter = sourceAdapter; }

	var sourceArray = sourceAdapter.isChildren(sourceTree);
	if (sourceArray) { sourceTree = listToNode(sourceTree, sourceAdapter); }

	if (createTree) { targetTree = targetAdapter.initRoot(); }
	else if (!targetTree) { targetTree = sourceTree; }
	else if (targetAdapter.isChildren(targetTree)) {
		targetTree = listToNode(targetTree, targetAdapter);
	}

	var tc = sourceAdapter.childCount(sourceTree);
	var nodeInput, nodeOutput, arrayOutput, j, jOffset = 0;
	var parentBreadcrumb = ctx.getBreadcrumb;
	var getBreadcrumbLast = function() {
		return {
			// node: tree[j],
			index: j,
			indexOffset: jOffset,
			indexTarget: j+jOffset,
			node: sourceAdapter.childIndexGet(sourceTree, j),
			parent: sourceTree,
			parentTarget: targetTree,
		};
	};
	var getBreadcrumb = function() {
		var par = parentBreadcrumb && parentBreadcrumb();
		var last = getBreadcrumbLast();
		return par ? par.concat( [last]) : [last];
	};
	for (j = 0; j < tc; j++) {
		nodeInput = sourceAdapter.childIndexGet(sourceTree, j);
		ctx.getBreadcrumbLast = getBreadcrumbLast;
		ctx.getBreadcrumb = getBreadcrumb;
		nodeOutput = plugin(nodeInput, sourceAdapter, ctx, targetAdapter, createTree, targetTree);
		arrayOutput = targetAdapter.toArray(nodeOutput);
		var nc = arrayOutput.length - 1;
		targetAdapter.childSplice(targetTree, j+jOffset, 1, arrayOutput);
		if (targetTree === sourceTree) {
			j += nc, tc += nc;
		} else {
			jOffset += nc;
		}
	}
	ctx.getBreadcrumb = parentBreadcrumb;
	if (sourceArray) {
		targetTree = targetAdapter.childrenGet(targetTree);
	}
	return targetTree;
}

function treeRender(sourceTree, sourceAdapter, ctx, plugins) {
	var assign;

	var pc = plugins.length, i;
	for (i = 0; i < pc; i++) {
		var plugin = plugins[i];
		var targetAdapter, createTree, targetTree;
		if (!(plugin instanceof Function)) {
			((assign = plugin, plugin = assign.plugin, targetAdapter = assign.targetAdapter, createTree = assign.createTree, targetTree = assign.targetTree));
		}
		sourceTree = treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree, targetTree);
		if (targetAdapter) { sourceAdapter = targetAdapter; }
	}
	return sourceTree;
}

function pluginListChildren(node, adapter, ctx, plugins) {
	if (!adapter.isText(node)) {
		var tree = adapter.childrenGet(node);
		if (tree) { adapter.childrenSet(node, treeRender(tree, adapter, ctx, plugins)); }
	}
	return node;
}

function pluginChildren(node, adapter, ctx, plugin, createTree) {
	if (!adapter.isText(node)) {
		var tree = adapter.childrenGet(node);
		if (tree) { adapter.childrenSet(node, treeRenderPlugin(tree, adapter, ctx, plugin, targetAdapter, createTree)); }
	}
	return node;
}

function pluginAttr(node, adapter, ctx, attrHandler, resultNode) {
	adapter.attrsEach(node, function (name, value, attr, index) {
		attrHandler(name, value, node, ctx, attr, index);
	});
	return resultNode ? resultNode(node, ctx) : node;
}

// plugin(nodeInput, sourceAdapter, ctx, targetAdapter, createTree);
function pluginConvertElement(node, adapter, ctx, targetAdapter, createTree) {
	var nodeTo;
	if (adapter.isText(node)) {
		nodeTo = targetAdapter.textNode(adapter.textValueGet(node));
	} else {
		nodeTo = targetAdapter.initName(adapter.nameGet(node));
		adapter.attrsEach(node, function (name, value, attr) {
			if (!attr) { attr = {name: name, value: value}; }
			targetAdapter.attrsAdd(nodeTo, attr);
		});
		// children = targetAdapter.childrenGet(nodeTo);
		treeRenderPlugin(node, adapter, ctx, pluginConvertElement, targetAdapter, createTree, nodeTo);
		// treeRenderPlugin(children, adapter, ctx, pluginConvertElement, targetAdapter, createTree);
		// treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree) {
	}
	return nodeTo;
}

function adapterPluginConvertElement(targetAdapter) {
	return plugin;
	function plugin(nodeInput, sourceAdapter, ctx, _, createTree) {
		return pluginConvertElement(nodeInput, sourceAdapter, ctx, targetAdapter, createTree);
	}
}

var treerender = /*#__PURE__*/Object.freeze({
__proto__: null,
listToNode: listToNode,
treeRenderPlugin: treeRenderPlugin,
treeRender: treeRender,
pluginListChildren: pluginListChildren,
pluginChildren: pluginChildren,
pluginAttr: pluginAttr,
pluginConvertElement: pluginConvertElement,
adapterPluginConvertElement: adapterPluginConvertElement
});

function ObjectAssign(target) {
	var arguments$1 = arguments;

	var hop = Object.prototype.hasOwnProperty;
	var argc = arguments.length;
	for (var i = 1; i < argc; i++) {
		var source = arguments$1[i];
		for (var key in source) {
			if (hop.call(source, key)) {
				target[key] = source[key];
			}
		}
	}
	return target;
}

var slice = Array.prototype.slice;

var events = {
	startTag: function(ev) {
		this.treeEvent('tagInit', null, ev);
	},
	tagName: function(ev) {
		if (ev.tag.close) {
			this.findAndCloseTag(ev);
		} else {
			this.openTag(ev);
		}
	},
	tagAttribute: function(ev) {
		this.element.attrsAdd(this.currentScope.tag, ev.attr, ev, this);
	},
	endTag: function(ev) {
		var breadcrumb = this.currentScope;
		var tagClose = ev.tag.close;
		var tagSelfClose = ev.tag.selfClose;
		if (!tagClose) {
			this.treeEvent('tagOpenEnd', null, ev);
			this.path.push(breadcrumb);
			this.element.childElement(breadcrumb.parentScope.tag, breadcrumb.tag, ev, this);
		}
		if (tagSelfClose) {
			this.findAndCloseTag(ev);
		}
		if (tagClose || tagSelfClose) {
			var tagOpen = this.closeTagMatch;
			// if tagOpen has no value, an error was already fired
			// in the function findAndCloseTag
			if (tagOpen) {
				var breadcrumbClose = tagOpen.match;
				this.treeEvent('tagCloseEnd', null, ev);
				this.path = this.path.slice(0, tagOpen.pathIndex);
				this.currentScope = breadcrumbClose.parentScope;
				this.closeTagMatch = null;
			}
		}
	},
	text: function(ev) {
		this.treeEvent('text', null, ev);
		this.element.childText(this.currentScope.tag, ev.text, ev, this);
	}
};

function noop(){}

function TreeError(message, code, state) {
	this.message = message;
	this.code = code;
	this.state = state;
	if (arguments.length > 2) { this.extras = slice.call(arguments, 2); }
}
TreeError.prototype = new Error;
TreeError.prototype.constructor = TreeError;
TreeError.prototype.name = 'TreeError';
TreeError.prototype.code = NaN;
TreeError.prototype.state = null;
TreeError.prototype.extras = null;

function TreeBuilder(opt) {
	if (opt instanceof Function) { opt = {event:opt}; }
	if (opt) { opt = ObjectAssign({}, TreeBuilder.optDefault, opt); }
	else { opt = ObjectAssign({}, TreeBuilder.optDefault); }
	this.eventFn = opt.event || noop;
	this.element = opt.element;
	this.tagVoidMap = opt.tagVoidMap;
	this.scopeNewChild(this.element.initRoot());
	this.root = this.currentScope;
	this.path = [];
	this.errors = [];
	this.parserEvent = this.parserEvent.bind(this);
}
TreeBuilder.optDefault = {
	element: null,
	tagVoidMap: null
};
TreeBuilder.prototype = {
	constructor: TreeBuilder,
	eventFn: null,
	root: null,
	currentScope: null,
	closeTagMatch: null,
	path: null,
	errors: null,
	tagVoidMap: null,
	parserEvent: function(ev) {
		var handler = this.events[ev.name];
		if (handler) {
			handler.call(this, ev);
		}
	},
	getParserState: function(ref) {
		var parser = ref.parser;
		var id = ref.id;

		return {
			chr: parser.c,
			pos: parser.pos,
			line: parser.line,
			column: parser.column,
			endPos: parser.endPos,
			endLine: parser.endLine,
			endColumn: parser.endColumn,
			currentTag: parser.currentTag,
			currentAttr: parser.currentAttr,
			eventId: id
		};
	},
	getEventObject: function(name, error, event) {
		var cs = this.currentScope;
		return {
			name: name,
			error: error,
			event: event,
			tag: cs.tag,
			text: event.text,
			parent: cs.parentScope,
			path: this.path,
			tagClose: this.closeTagMatch,
			builder: this
		};
	},
	treeEvent: function(name, error, ev) {
		this.eventFn(this.getEventObject(name, error, ev));
	},
	scopeNewChild: function(child) {
		var parent = this.currentScope;
		this.currentScope = {
			tag: child,
			parentScope: parent
		};
	},
	openTag: function(ev) {
		this.scopeNewChild(this.element.initName(ev.tag.name, ev, this));
		this.treeEvent('tagOpenStart', null, ev);
	},
	unclosedTagChildren: function(tag, index, ev) {
		var name = this.element.nameGet(tag, ev.event, this);
		name = String(name).toLowerCase();
		if (this.tagVoidMap[name]) { return 0; }
	},	
	resolveUnclosedTags: function(unclosed, ev) {
		var uclen = unclosed.length;
		while (uclen) {
			var ucIndex = uclen - 1;
			var uc = unclosed[ucIndex];
			var ucTag = uc.tag;
			var ucEv = this.getEventObject('resolveUnclosedTags', null, ev);
			var ucCount = this.unclosedTagChildren(ucTag, ucIndex, ucEv);
			if (ucCount === void 0) { break; }
			var ucParent = uc.parentScope.tag;
			var ucSiblings = ucParent.children;
			var ucSelfPos = ucSiblings.indexOf(ucTag);
			if (-1 === ucSelfPos) {
				err = new TreeError(
					'Could not find child tag '+JSON.stringify(ucTag.name)+
					' in parent tag '+JSON.stringify(ucParent.name)+
					' list of '+ucSiblings.length+' children',
					102
				);
				this.errors.push(err);
				this.treeEvent('error', err, ev);
				break;
			} else {
				var ucRemCount = ucTag.children.length - ucCount;
				var ucRem = this.element.childSplice(ucTag, ucCount, ucRemCount, [], ev, this);
				this.element.childSplice(ucParent, ucSelfPos+1, 0, ucRem, ev, this);
				this.treeEvent('tagCloseStart', null, ev);
				this.treeEvent('tagCloseEnd', null, ev);
				unclosed.pop();
				uclen--;
				this.path.pop();
				this.currentScope = this.currentScope.parentScope;
			}
		}
		return unclosed;
	},
	findAndCloseTag: function(ev) {
		var err;
		var streamTag = ev.tag;
		var tagOpen = this.findTagOpen(streamTag, ev);
		this.closeTagMatch = tagOpen;
		if (tagOpen) {
			var breadcrumb = tagOpen.match;
			if (streamTag.selfClose) { breadcrumb.tag.selfClose = true; }
			var unclosed = this.resolveUnclosedTags(tagOpen.unclosedTags, ev);
			var uclen = unclosed.length;
			if (uclen) {
				err = new TreeError(
					(streamTag.selfClose ? 'Self closing tag ' : 'Close tag ')+
					JSON.stringify(streamTag.name)+' with '+uclen+' open tags',
					streamTag.selfClose ? 101 : 100
				);
				this.errors.push(err);
				this.treeEvent('unclosedTags', err, ev);
			}
			this.treeEvent('tagCloseStart', null, ev);
		} else {
			err = new TreeError(
				'Close tag '+JSON.stringify(streamTag.name)+' without opening tag',
				102
			);
			this.errors.push(err);
			this.treeEvent('unopenedTag', err, ev);
		}
	},
	findTagOpen: function(streamTag, ev) {
		var p = this.path;
		var match, i;
		for (i = p.length - 1; 0 <= i; i--) {
			var ptag = p[i];
			if (this.element.nameGet(ptag.tag, ev, this) === streamTag.name) {
				match = ptag;
				break;
			}
		}
		if (match) { return {
			match: match,
			pathIndex: i,
			unclosedTags: p.slice(i+1)
		}; }
	},
	events: events
};

TreeBuilder.TreeError = TreeError;

function arrayFrom(val) {
	var slice = Array.prototype.slice;
	switch (true) {
		case val instanceof Array: return val;
		case void 0 == val: return [];
		case isCollection(val): return slice.call(val);
		default: return [val];
	}
}

function arrayConcat(val) {
	switch (true) {
		case val instanceof Array: return val;
		case void 0 == val: return [];
		default: return [val];
	}
}

function isCollection(obj) {
	var propEnum = Object.prototype.propertyIsEnumerable;
	var len, intLen;
	return obj &&
		'length' in obj &&
		!propEnum.call(obj, 'length') &&
		(len = obj.length,
			intLen = parseInt(len),
			intLen === len && intLen >= 0);
}

function elementDefault (ref) {
	if ( ref === void 0 ) ref = {};
	var keyName = ref.keyName; if ( keyName === void 0 ) keyName = 'name';
	var keyAttrs = ref.keyAttrs; if ( keyAttrs === void 0 ) keyAttrs = 'attrs';
	var keyChildren = ref.keyChildren; if ( keyChildren === void 0 ) keyChildren = 'children';
	var keyText = ref.keyText; if ( keyText === void 0 ) keyText = 'text';
	var textName = ref.textName; if ( textName === void 0 ) textName = '#text';
	var rootName = ref.rootName; if ( rootName === void 0 ) rootName = '#document-fragment';
	var commentName = ref.commentName; if ( commentName === void 0 ) commentName = '#comment';

	var initName = function (name) {
		var obj;

		return (( obj = {}, obj[keyName] = name, obj[keyAttrs] = [], obj[keyChildren] = [], obj[keyText] = null, obj ));
	};
	var child = function (el, child) { return void el[keyChildren].push(child); };
	var textNode = function (text) {
		var obj;

		return (( obj = {}, obj[keyName] = textName, obj[keyText] = text, obj ));
	};
	return {
		isText: function (el) { return textName === el[keyName]; },
		isFragment: function (el) { return el[keyName] === rootName; },
		isComment: function (el) { return el[keyName] === commentName; },
		isChildren: function (ch) { return ch instanceof Array; },
		initRoot: function () { return initName(rootName); },
		initName: initName,
		initComment: function (text) {
			if ( text === void 0 ) text = '';

			return (el = initName(commentName), el[keyText] = text, el);
	},
		nameGet: function (el) { return el[keyName]; },
		textNode: textNode,
		textValueGet: function (el) { return el[keyText]; },
		textValueSet: function (el, text) { return el[keyText] = text; },
		attrsAdd: function (el, attr) { return void el[keyAttrs].push(attr); },
		attrsEach: function (el, handler) {
			var list = el[keyAttrs];
			if (!list) { console.log('ElementDefault attrs not found', el, keyAttrs); }
			var count = list && list.length || 0;
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			for (var i = 0; i < count; i++) {
				var a = list[i];
				var ret = handler.call(ctx, a.name, a.value, a, i);
				if (ret & ctx._remove) { list.splice(i, 1), i--, count--; }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: function (el) { return el[keyChildren].length; },
		childIndexGet: function (el, index) { return el[keyChildren][index]; },
		childSplice: function (el, index, remove, add) {
			var ref;

			return (ref = el[keyChildren]).splice.apply(ref, [ index, remove ].concat( (add || []) ));
	},
		childrenGet: function (el) { return el[keyChildren]; },
		childrenSet: function (el, children) { return el[keyChildren] = children; },
		toArray: arrayConcat,
	};
}

function dom (apiDom) {
	var child = function (el, child) { return void el.appendChild(child); };
	var childCount = function (el) { return el.childNodes.length; };
	var childIndexGet = function (el, index) { return el.childNodes[index]; };
	var childSplice = function (el, index, remove, add) {
		//splice.apply(el.childNodes, [index, remove, ...(add || [])]);
		var ec = childCount(el);
		while (remove > 0 && ec > index) {
			el.removeChild(childIndexGet(el, index));
			remove--, ec--;
		}
		var i, c = add.length;
		var temp = []; // Live NodeList changes as we reassign nodes
		for (i = 0; i < c; i++) {
			temp.push(add[i]);
		}
		add = temp;
		for (i = 0; i < c; i++) {
			el.insertBefore(add[i], childIndexGet(el, index + i));
		}
	};
	var textNode = function (text) { return apiDom.createTextNode(text); };
	var isFragment = function (el) { return el.nodeType === apiDom.DOCUMENT_FRAGMENT_NODE; };
	var isComment = function (el) { return el.nodeType === apiDom.COMMENT_NODE; };
	var isChildren = function (ch) { return ch && ch.constructor === apiDom.childNodes.constructor; };
	// const splice = Array.prototype.splice;
	return {
		isText: function (el) { return el.nodeType === apiDom.TEXT_NODE; },
		isFragment: isFragment,
		isComment: isComment,
		isChildren: isChildren,
		initRoot: function () { return apiDom.createDocumentFragment(); },
		initName: function (name) { return apiDom.createElement(name); },
		initComment: function (text) {
			if ( text === void 0 ) text = '';

			return apiDom.createComment(text);
	},
		nameGet: function (el) { return el.nodeName; },
		textNode: textNode,
		textValueGet: function (el) { return el.nodeValue; },
		textValueSet: function (el, text) { return void (el.nodeValue = text); },
		attrsAdd: function (el, attr) { return el.setAttribute(attr.name, attr.value); },
		attrsEach: function (el, handler) {
			var list = el.attributes;
			if (!list) { console.log('ElementDom attrs not found', el); }
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			var count = list.length;
			for (var i = 0; i < count; i++) {
				var a = list[i];
				var ret = handler.call(ctx, a.name, a.value, a, i);
				if (ret & ctx._remove) { el.removeAttribute(a.name); }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: childCount,
		childIndexGet: function (el, index) { return el.childNodes[index]; },
		childSplice: childSplice,
		childrenGet: function (el) { return el.childNodes; },
		childrenSet: function (el, children) { return childSplice(el, 0, childCount(el), children); },
		toArray: function (el) {
			return !el ? [] :
				isFragment(el) ? arrayFrom(el.childNodes) :
				el.constructor === apiDom.childNodes.constructor ? arrayFrom(el) :
				arrayConcat(el);
		}
	};
}

var initName = function (name) { return ({
	_name: name,
	_class: null,
	_id: null,
	sel: name,
	data: {},
	text: void 0, // Snabbdom requires undefined
	children: []
}); };
var textNode = function (text) { return ({text:text}); };
var child = function (el, child) { return void el.children.push(child); };
var attrIdSel = function (attr, el) {
	if (attr.name === 'id') {
		el._id = attr.value;
		attrUpdateSel(el);
		return true;
	}
};
var attrClassSel = function (attr, el) {
	if (attr.name === 'class') {
		el._class = attr.value;
		attrUpdateSel(el);
		return true;
	}
};
var attrUpdateSel = function (el) {
	var c = el._class;
	var i = el._id;
	el.sel = el._name +
		(i ? '#' + i : '') +
		(c ? '.' + c.replace(/\s+/g,'.') : '');
};
var attrObject = function (attr, el, key) {
	var ea = el.data[key];
	if (!ea) { el.data[key] = ea = {}; }
	ea[attr.name] = attr.value;
	return true;
};
var attrHandlerDefault = function (el, attr) {
	return attrIdSel(attr, el)
		|| attrClassSel(attr, el)
		|| attrObject(attr, el, 'attrs');
};
var attrFn = {
	attrIdSel: attrIdSel,
	attrClassSel: attrClassSel,
	attrUpdateSel: attrUpdateSel,
	attrObject: attrObject,
	attrHandlerDefault: attrHandlerDefault
};
var refFragment = {};
var nameComment = '!';

function snabbdom (ref) {
	if ( ref === void 0 ) ref = {};
	var attrHandler = ref.attrHandler;

	attrHandler = (attrHandler || attrHandlerDefault).bind(attrFn);
	return {
		isText: function (el) { return !el.sel && 'string' === typeof el.text; },
		isFragment: function (el) { return el._name === refFragment; },
		isComment: function (el) { return el._name === nameComment; },
		isChildren: function (ch) { return ch instanceof Array; },
		initRoot: function () { return initName(refFragment); },
		initName: initName,
		initComment: function (text) {
		if ( text === void 0 ) text = '';
var el = initName(nameComment); el.text = text; return el},
		nameGet: function (el) { return el._name; },
		textNode: textNode,
		textValueGet: function (el) { return el.text; },
		textValueSet: function (el, text) { return void (el.text = text); },
		attrsAdd: attrHandler,
		attrsEach: function (el, handler) {
			if (el && !el.data) { console.log('ElementSnabbdom data not found', el); }
			var map = el.data.attrs;
			var hop = Object.prototype.hasOwnProperty;
			if (!map) { return; }
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			var keys = [], k;
			for (k in map) { if (hop.call(map, k)) { keys.push(k); } }
			for (var i = 0, count = keys.length; i < count; i++) {
				k = keys[i];
				var ret = handler.call(ctx, k, map[k], null, i);
				if (ret & ctx._remove) { delete map[k]; }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: function (el) { return el.children.length; },
		childIndexGet: function (el, index) { return el.children[index]; },
		childSplice: function (el, index, remove, add) {
			var ref;

			return (ref = el.children).splice.apply(ref, [ index, remove ].concat( (add || []) ));
	},
		childrenGet: function (el) { return el.children; },
		childrenSet: function (el, children) { return el.children = children; },
		toArray: arrayConcat,
	};
}

var htmlVoidTagMap = {
	area:true,// clickable area inside an image map.
	base:true,// new root or base path for relative paths.
	basefont:true,
	bgsound:true,
	br:true,// single line break in a paragraph.
	col:true,
	command:true,
	embed:true,// embeds things like flash.
	frame:true,
	hr:true,// change in type of content.
	iframe:true,// frame inside the page.
	image:true,
	img:true,// image.
	input:true,// input field.
	isindex:true,
	keygen:true,
	link:true,// reference an external resource.
	menuitem:true,
	meta:true,// metadata about the page.
	nextid:true,
	param:true,// parameter for an object.
	source:true,// include media for video or audio.
	track:true,// text track for video or audio.
	wbr:true// word break opportunity
};

var st_TEXT = {name:'TEXT'};
var st_TAG_START = {name:'TAG_START'};
var st_TAG_NAME = {name:'TAG_NAME'};
var st_TAG_NAMED = {name:'TAG_NAMED'};
var st_TAG_NAMED_SELFCLOSE = {name:'TAG_NAMED_SELFCLOSE'};
var st_ATTR_NAME = {name:'ATTR_NAME'};
var st_ATTR_NAMED = {name:'ATTR_NAMED'};
var st_ATTR_VALUE_START = {name:'ATTR_VALUE_START'};
var st_ATTR_VALUE_QUOTED = {name:'ATTR_VALUE_QUOTED'};
var st_ATTR_VALUE_RAW = {name:'ATTR_VALUE_RAW'};
var st_ATTR_VALUE_CUSTOM = {name:'ATTR_VALUE_CUSTOM'};
var st_INSTRUCTION = {name:'INSTRUCTION'};
var st_DECLARATION = {name:'DECLARATION'};
var st_COMMENT = {name:'COMMENT'};
var st_CDATA = {name:'CDATA'};

var ev_text = {name:'text'};
var ev_startTag = {name:'startTag'};
var ev_endTag = {name:'endTag'};
var ev_startInstruction = {name:'startInstruction'};
var ev_endInstruction = {name:'endInstruction'};
var ev_startDeclaration = {name:'startDeclaration'};
var ev_endDeclaration = {name:'endDeclaration'};
var ev_startComment = {name:'startComment'};
var ev_endComment = {name:'endComment'};
var ev_startCdata = {name:'startCdata'};
var ev_endCdata = {name:'endCdata'};
var ev_tagName = {name:'tagName'};
var ev_tagAttribute = {name:'tagAttribute'};

var reSpace = /\s/;
var reCdata = /^\[CDATA\[$/i;
var lenCdata = 7;
var echo = function (x) { return x; };
var nop = function () {};

function XMLParser(opt) {
	if (opt instanceof Function) {
		this.event = opt;
		this.opt = { event: opt };
	} else {
		this.event = opt.event;
		this.opt = opt;
	}
	this.decodeString = opt.decodeString || echo;
	this.decodeText = opt.decodeText || this.decodeString;
	this.decodeTagName = opt.decodeTagName || this.decodeString;
	this.decodeAttrName = opt.decodeAttrName || this.decodeString;
	this.decodeAttrValue = opt.decodeAttrValue || this.decodeString;
	var cp = opt.customParser;
	this.getCustomParser = ObjectAssign({}, {attrValue: nop},
		cp);
	this.currentCustomParser = {
		attrValue: null
	};
	this.state = st_TEXT;
	this.buffer = '';
}

XMLParser.prototype = {
	constructor: XMLParser,
	opt: null,
	state: null,
	c: null,
	pos: 0,
	line: 0,
	column: 0,
	endPos: 0,
	endLine: 0,
	endColumn: 0,
	tagPath: null,
	currentTag: null,
	currentAttr: null,
	decodeString: null,
	decodeText: null,
	decodeTagName: null,
	decodeAttrName: null,
	decodeAttrValue: null,
	customParser: null,
	currentCustomParser: null,
	buffer: null,
	createAttr: function() {
		return {
			startSpace: null,
			name: null,
			eqSpace: null,
			value: null,
			valueSpace: null,
			quotes: null
		};
	},
	decodeTag: function(ctag, ev) {
		var name = ctag.name;
		if ('string' === typeof name) {
			name = this.decodeTagName(name, ev);
		}
		return ObjectAssign({}, ctag, {name: name});
	},
	decodeAttr: function(cattr, ev) {
		var name = cattr.name;
		var value = cattr.value;
		if ('string' === typeof name) {
			name = this.decodeAttrName(name, ev);
		}
		if ('string' === typeof value) {
			value = this.decodeAttrValue(value, ev);
		}
		return ObjectAssign({}, cattr, {name: name, value: value});
	},
	write: function(text, final) {
		var assign;

		function getEndChars(n) {
			var j = i + 1;
			if (j >= n) {
				return text.substr(j - n, n);
			} else {
				var blen = buf.length;
				if (j + blen >= n) {
					return buf.substr(blen - (n - j), n - j).concat(text.substr(0, j))
				} else {
					throw new Error('Requested '+n+' end characters but there are '+j+' in text beginning and '+blen+' in buffer');
				}
			}
		}
		function event(ev, id) {
			ev = {
				name: ev.name,
				id: id,
				attr: cattr,
				tag: ctag,
				text: buf,
				parser: self,
				customParser: self.currentCustomParser
			};
			if (ctag) { ev.tag = self.decodeTag(ctag, ev); }
			if (cattr) { ev.attr = self.decodeAttr(cattr, ev); }
			if (buf) { ev.text = self.decodeText(buf, ev); }
			eventFn(ev);
		}
		function eventEndTag(id, ev_custom) {
			buf = '';
			state = st_TEXT;
			event(ev_custom || ev_endTag, id);
			self.currentTag = ctag = null;
		}
		function eventTagAttr(id) {
			event(ev_tagAttribute, id);
			self.currentAttr = cattr = null;
		}
		function eventTagNameSlash(id) {
			ctag.name = tagNameSlash;
			event(ev_tagName, id);
			tagNameSlash = void 0;
		}
		var self = this;
		text = text || '';
		var tlen = text.length;
		var eventFn = this.event;
		var state = this.state;
		var ctag = this.currentTag;
		var cattr = this.currentAttr;
		var buf = this.buffer;
		var pos = this.pos;
		var line = this.line;
		var column = this.column;
		var attrQuoteChar = this.attrQuoteChar;
		var tagBeforeClose = this.tagBeforeClose;
		var tagNameSlash = this.tagNameSlash;
		var ccp = this.currentCustomParser;
		for (var i = 0; i < tlen; i++) {
			var c = this.c = text[i];
			pos++;
			if (c === '\n') {
				line++;
				column = 0;
			} else {
				column++;
			}
			this.endPos = pos;
			this.endLine = line;
			this.endColumn = column;
			switch (state) {
				case st_TEXT:
					switch (c) {
						case '<':
							if (buf) {
								event(ev_text, '10');
								buf = '';
							}
							state = st_TAG_START;
							this.currentTag = ctag = {};
							event(ev_startTag, '20');
							break;
						default:
							buf += c;
					}
					break;
				case st_TAG_START:
					switch (c) {
						case '?':
							if (buf) { ctag.startSpace = buf; }
							ctag.instruction = true;
							buf = '';
							state = st_INSTRUCTION;
							event(ev_startInstruction, '30');
							break;
						case '!':
							if (buf) { ctag.startSpace = buf; }
							ctag.declaration = true;
							buf = '';
							state = st_DECLARATION;
							event(ev_startDeclaration, '40');
							break;
						case '>':
							ctag.empty = true;
							if (buf) {
								if (ctag.close) {
									ctag.endSpace = buf;
								} else {
									ctag.startSpace = buf;
								}
								buf = '';
							}
							event(ev_tagName, '45');
							eventEndTag('50');
							break;
						case '/':
							ctag.close = true;
							if (buf) { ctag.startSpace = buf; }
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (!buf) ; else if (ctag.close) {
									ctag.startCloseSpace = buf;
								} else {
									ctag.startSpace = buf;
								}
								state = st_TAG_NAME;
								buf = c;
							}
					}
					break;
				case st_INSTRUCTION:
					switch (c) {
						case '>':
							ctag.text = buf;
							eventEndTag('60', ev_endInstruction);
							break;
						default:
							buf += c;
					}
					break;
				case st_DECLARATION:
					switch (c) {
						case '>':
							ctag.text = buf;
							eventEndTag('70', ev_endDeclaration);
							break;
						case '-':
							if (getEndChars(2) === '--') {
								ctag.beforeComment = buf.substr(0, buf.length - 1);
								ctag.comment = true;
								buf = '';
								state = st_COMMENT;
								event(ev_startComment, '80');
								break;
							}
						default:
							buf += c;
							if (buf.length === lenCdata && reCdata.test(buf)) {
								ctag.cdata = true;
								buf = '';
								state = st_CDATA;
								event(ev_startCdata, '90');
								break;
							}
					}
					break;
				case st_COMMENT:
					switch (c) {
						case '-':
							if (getEndChars(2) === '--') {
								ctag.textComment = buf.substr(0, buf.length - 1);
								buf = '';
								state = st_DECLARATION;
								event(ev_endComment, '100');
								break;
							}
						default:
							buf += c;
					}
					break;
				case st_CDATA:
					switch (c) {
						case ']':
							if (getEndChars(2) === ']]') {
								ctag.textCdata = buf.substr(0, buf.length - 1);
								buf = '';
								state = st_DECLARATION;
								event(ev_endCdata, '110');
								break;
							}
						default:
							buf += c;
					}
					break;
				case st_TAG_NAME:
					switch (c) {
						case '>':
							if (tagBeforeClose) {
								ctag.name = tagBeforeClose;
								ctag.selfClose = true;
								this.tagBeforeClose = tagBeforeClose = void 0;
							} else {
								ctag.name = buf;
							}
							event(ev_tagName, '120');
							eventEndTag('130');
							break;
						case '/':
							if (ctag.close) {
								buf += c;
							} else {
								this.tagBeforeClose = tagBeforeClose = buf;
								buf = '';
							}
							break;
						default:
							if (reSpace.test(c)) {
								if (tagBeforeClose) {
									this.tagNameSlash = tagNameSlash = tagBeforeClose;
									this.tagBeforeClose = tagBeforeClose = void 0;
									state = st_TAG_NAMED_SELFCLOSE;
								} else {
									ctag.name = buf;
									state = st_TAG_NAMED;
									event(ev_tagName, '140');
								}
								buf = c;
							} else if (tagBeforeClose) {
								buf = tagBeforeClose+'/'+c;
								this.tagBeforeClose = tagBeforeClose = void 0;
							} else {
								buf += c;
							}
					}
					break;
				case st_TAG_NAMED:
					switch (c) {
						case '>':
							if (buf) { ctag.endSpace = buf; }
							eventEndTag('150');
							break;
						case '/':
							this.tagBeforeClose = tagBeforeClose = buf;
							state = st_TAG_NAMED_SELFCLOSE;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_TAG_NAMED_SELFCLOSE:
					switch (c) {
						case '>':
							if (tagBeforeClose) { ctag.selfCloseSpace = tagBeforeClose; }
							if (buf) { ctag.endSpace = buf; }
							if (tagNameSlash) { eventTagNameSlash('155'); }
							ctag.selfClose = true;
							this.tagBeforeClose = tagBeforeClose = void 0;
							eventEndTag('160');
							break;
						case '/':
							if (tagNameSlash) {
								this.tagNameSlash = tagNameSlash += '/';
								eventTagNameSlash('165');
							} else if (tagBeforeClose) {
								this.currentAttr = cattr = this.createAttr();
								cattr.startSpace = tagBeforeClose;
								cattr.name = c;
								eventTagAttr('170');
							}
							this.tagBeforeClose = tagBeforeClose = buf;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (tagNameSlash) {
									this.tagNameSlash = tagNameSlash += '/';
									eventTagNameSlash('175');
								}
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_ATTR_NAME:
					switch (c) {
						case '>':
							cattr.name = buf;
							eventTagAttr('180');
							eventEndTag('190');
							break;
						case '=':
							cattr.name = buf;
							state = st_ATTR_VALUE_START;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								cattr.name = buf;
								state = st_ATTR_NAMED;
								buf = c;
							} else {
								buf += c;
							}
					}
					break;
				case st_ATTR_NAMED:
					switch (c) {
						case '>':
							ctag.endSpace = buf;
							eventTagAttr('200');
							eventEndTag('210');
							break;
						case '=':
							if (buf) { cattr.eqSpace = buf; }
							cattr.eq = true;
							state = st_ATTR_VALUE_START;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (cattr) {
									eventTagAttr('220');
								}
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_ATTR_VALUE_START:
					switch (c) {
						case '>':
							if (buf) { ctag.endSpace = buf; }
							eventTagAttr('230');
							eventEndTag('240');
							break;
						case '"':
						case '\'':
							cattr.valueSpace = buf;
							this.attrQuoteChar = cattr.quotes = attrQuoteChar = c;
							state = st_ATTR_VALUE_QUOTED;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (buf) { cattr.valueSpace = buf; }
								ccp.attrValue = this.getCustomParser.attrValue(c, cattr);
								if (ccp.attrValue) {
									state = st_ATTR_VALUE_CUSTOM;
									buf = '';
								} else {
									state = st_ATTR_VALUE_RAW;
									buf = c;
								}
							}
					}
					break;
				case st_ATTR_VALUE_RAW:
					switch (c) {
						case '>':
							cattr.value = buf;
							eventTagAttr('250');
							eventEndTag('260');
							break;
						default:
							if (reSpace.test(c)) {
								cattr.value = buf;
								state = st_TAG_NAMED;
								eventTagAttr('270');
								buf = c;
							} else {
								buf += c;
							}
					}
					break;
				case st_ATTR_VALUE_QUOTED:
					switch (c) {
						case attrQuoteChar:
							this.attrQuoteChar = attrQuoteChar = void 0;
							cattr.value = buf;
							buf = '';
							state = st_TAG_NAMED;
							eventTagAttr('280');
							break;
						default:
							buf += c;
					}
					break;
				case st_ATTR_VALUE_CUSTOM:
					buf = ccp.attrValue.push(c, {attr: cattr, i: i, pos: pos, line: line, column: column});
					if (buf != void 0) {
						state = st_TAG_NAMED;
						ccp.attrValue = null;
						((assign = buf, cattr = assign.attr, i = assign.i, pos = assign.pos, line = assign.line, column = assign.column));
						buf = '';
						if (cattr) { eventTagAttr('290'); }
					}
					break;
			}
			this.pos = this.endPos;
			this.line = this.endLine;
			this.column = this.endColumn;
		}
		this.c = void 0;
		this.state = state;
		if (final && state === st_TEXT && buf) {
			event(ev_text, '290');
			buf = '';
		}
		this.buffer = buf;
	},
	end: function(text) {
		this.write(text, true);
	}
};

var echo$1 = function (x) { return x; };

function Printer(opt) {
	this.adapter = opt && opt.adapter;
	this.indent = opt && opt.indent;
	this.encodeString = opt && opt.encodeString || echo$1;
	this.encodeText = opt && opt.encodeText || this.encodeString;
	this.encodeTagName = opt && opt.encodeTagName || this.encodeString;
	this.encodeAttrName = opt && opt.encodeAttrName || this.encodeString;
	this.encodeAttrValue = opt && opt.encodeAttrValue || this.encodeString;
	this.selfCloseString = opt && opt.selfCloseString || this.selfCloseString;
	this.newLine = opt && opt.newLine || this.newLine;
}
Printer.prototype = {
	constructor: Printer,
	elAdapter: null,
	indent: null,
	encodeString: null,
	encodeText: null,
	encodeTagName: null,
	encodeAttrName: null,
	encodeAttrValue: null,
	selfCloseString: ' /',
	newLine: '\n',
	log: function() {},
	repeat: function(n, c) {
		var s = '';
		while (n > 0) {
			s += c;
			n--;
		}
		return s;
	},
	printIndent: function(level) {
		var indent = this.indent;
		if (indent instanceof Function) {
			return indent(level);
		}
		if (null == indent) {
			indent = '\t';
		} else if (+indent === indent) {
			indent = this.repeat(indent, ' ');
		}
		return this.repeat(level, indent);
	},
	isVoidTag: function(node) {
		var name = this.elAdapter.nameGet(node);
		name = String(name || '').toLowerCase();
		return htmlVoidTagMap[name];
	},
	printAttr: function(name, value) {
		var attr = this.encodeAttrName(name);
		if (null != value) {
			attr += '="'+this.encodeAttrValue(value)+'"';
		}
		return attr;
	},
	printTagOpen: function(node, isSelfClose) {
		var self = this;
		var name = this.elAdapter.nameGet(node);
		var tag = '<' + this.encodeTagName(name);
		this.elAdapter.attrsEach(node, function(name, value) {
			tag += ' '+self.printAttr(name, value);
		});
		tag += (isSelfClose ? this.selfCloseString : '') + '>';
		return tag;
	},
	printTagClose: function(node) {
		var name = this.elAdapter.nameGet(node);
		var tag = '</' + this.encodeTagName(name) + '>';
		return tag;
	},
	printTag: function(node, level) {
		var nl = this.newLine;
		var nc = this.elAdapter.childCount(node);
		var sc = 0 == nc && this.isVoidTag(node);
		var out = this.printIndent(level);
		out += this.printTagOpen(node, sc);
		if (nc > 0) {
			out += nl + this.print(this.elAdapter.childrenGet(node), level+1);
			out += this.printIndent(level);
		}
		if (!sc) {
			out += this.printTagClose(node);
		}
		out += nl;
		return out;
	},
	textSplitLines: function(text) {
		return text.split(/\r\n|\n|\r/g);
	},
	textTrim: function(text) {
		return text.replace(/^[ \t\r\n]+|[ \t]+(?=[ \t])|[ \t\r\n]+$/g, '');
	},
	printTextLine: function(text, level) {
		var trim = this.textTrim(text);
		return trim ? this.printIndent(level) + trim : trim;
	},
	printText: function(ftext, level) {
		var text = this.textTrim(ftext);
		text = this.textSplitLines(text);
		var c = text.length;
		var nl = this.newLine;
		var s = '';
		for (var i = 0; i < c; i++) {
			var l = this.printTextLine(text[i], level);
			s += l ? l + nl : '';
		}
		return s;
	},
	print: function(tree, level) {
		var tc = tree.length;
		var out = '';
		for (var i = 0; i < tc; i++) {
			var node = tree[i];
			if (this.elAdapter.isText(node)) {
				out += this.printText(this.elAdapter.textValueGet(node), level);
			} else {
				out += this.printTag(node, level);
			}
		}
		return out;
	}
};

function getPosStart(parser) {
	return {
		byte: parser.pos,
		line: parser.line,
		column: parser.column
	};
}
function getPosEnd(parser) {
	return {
		byte: parser.endPos,
		line: parser.endLine,
		column: parser.endColumn
	};
}

var statsEventMap = {
	tagInit: function statsTagInit(ev) {
		this.statsLastTagInit = {
			tag: ev.event.tag,
			pos: getPosStart(ev.event.parser)
		};
	},
	tagOpenStart: function statsTagOpenStart(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		var sl = this.statsLastTagInit;
		var streamTag = ev.event.tag;
		s.openSource = streamTag;
		s.openStart = (sl && sl.tag === streamTag && sl.pos) || getPosStart(ev.event.parser);
	},
	tagOpenEnd: function statsTagOpenEnd(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		s.openEnd = getPosEnd(ev.event.parser);
		this.statsLastTagInit = null;
	},
	tagCloseStart: function statsTagCloseStart(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		var sl = this.statsLastTagInit;
		var streamTag = ev.event.tag;
		s.closeSource = streamTag;
		s.closeStart = (sl && sl.tag === streamTag && sl.pos) || getPosStart(ev.event.parser);
	},
	tagCloseEnd: function statsTagCloseEnd(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		s.closeEnd = getPosEnd(ev.event.parser);
		this.statsLastTagInit = null;
	}
};
function statsEvent(ev) {
	var fn = statsEventMap[ev.name];
	if (fn) { fn.call(this, ev); }
}

function HTMLTypeset() {
	this.errors = [];
	this.text = [];
	this.rootStyle = {
		fontFace: 'helvetica',
		fontSize: 10,
		fontBold: false,
		fontItalic: false,
		lineLeading: 15,
		paragraphLeading: 25,
		tolerance: void 0
	};
	this.scopeReset();
	this.treeEvent = this.treeEvent.bind(this);
}
HTMLTypeset.prototype = {
	errors: null,
	text: null,
	currentParagraph: null,
	currentScope: null,
	pathScope: null,
	currentSpan: null,
	rootStyle: null,
	reInline: /^em$|^span$|^strong$|^b$|^i$/i,
	reBold: /^strong$|^b$/i,
	reItalic: /^em$|^i$/i,
	ensureParagraph: function () {
		if (this.currentParagraph) { return; }
		var s = this.currentScope || this.rootStyle;
		this.currentParagraph = {
			leading: s.lineLeading,
			paragraphLeading: s.paragraphLeading,
			tolerance: s.tolerance,
			spans: []
		};
	},
	pushParagraph: function () {
		if (!this.currentParagraph) { return; }
		this.pushSpan();
		this.text.push(this.currentParagraph);
		this.currentParagraph = void 0;
	},
	ensureSpan: function () {
		this.ensureParagraph();
		if (this.currentSpan) { return; }
		var cs = this.currentScope;
		this.currentSpan = {
			fontFamily: [
				cs.fontFace,
				(cs.fontBold || cs.fontItalic)
					? (cs.fontBold ? 'bold' : '') + (cs.fontItalic ? 'italic' : '')
					: 'normal'
			].join('\t'),
			fontSize: cs.fontSize,
			text: ''
		};
	},
	normalizeWhitespace: function(str) {
		return str.replace(/[ \r\n\t]+/g,' ');
	},
	decodeString: function(str) {
		return str;
		// this function is meant to be replaced
		// with a function to decode html entities
	},
	processTextCommit: function(text) {
		return this.decodeString(this.normalizeWhitespace(text));
	},
	pushSpan: function () {
		if (!this.currentSpan) { return; }
		var sText = this.currentSpan.text;
		if (sText) {
			this.currentSpan.text = this.processTextCommit(sText);
			this.currentParagraph.spans.push(this.currentSpan);
		}
		this.currentSpan = void 0;
	},
	scopeParent: function () {
		var cs = this.currentScope;
		return cs && cs.parent || this.rootStyle;
	},
	scopePush: function (tag, style) {
		var scope = this.currentScope || this.rootStyle;
		this.currentScope = ObjectAssign({}, scope,
			style,
			{tagOpen: tag,
			tagClose: null,
			parent: this.currentScope});
		this.pathScope.push(this.currentScope);
	},
	scopePop: function (tag) {
		this.currentScope.tagClose = tag;
		this.currentScope = this.currentScope.parent;
		this.pathScope.pop();
	},
	scopeReset: function () {
		this.currentScope = void 0;
		this.pathScope = [];
	},
	treeTagOpen: function(tagName, tag) {
		var style = {};
		if (this.reBold.test(tagName)) { style.fontBold = true; }
		if (this.reItalic.test(tagName)) { style.fontItalic = true; }
		this.scopePush(tag || tagName, style);
		if (this.reInline.test(tagName)) {
			this.pushSpan();
		} else {
			this.pushParagraph();
		}
	},
	treeTagClose: function(tagName, tag) {
		this.scopePop(tag);
		if (this.reInline.test(tagName)) {
			this.pushSpan();
		} else {
			this.pushParagraph();
		}
	},
	processTextInput: function(text) {
		return text;
	},
	treeText: function(text) {
		this.ensureSpan();
		this.currentSpan.text += this.processTextInput(text);
	},
	treeOnError: function(ev) {
		this.errors.push(ev);
	},
	treeEvent: function(ev) {
		var evName = ev.name;
		var tag = ev.tag;
		var tagName = tag && tag.name;
		if (ev.error) { this.treeOnError(ev); }
		switch (evName) {
			case 'tagOpenEnd': this.treeTagOpen(tagName, tag); break;
			case 'tagCloseEnd': this.treeTagClose(tagName, tag); break;
			case 'text': this.treeText(ev.text); break;
		}
	},
	getResult: function() {
		this.pushParagraph();
		return {
			errors: this.errors,
			result: this.text
		};
	}
};

TreeBuilder.optDefault.element = elementDefault;
TreeBuilder.optDefault.tagVoidMap = htmlVoidTagMap;

exports.HTMLTypeset = HTMLTypeset;
exports.Printer = Printer;
exports.TreeBuilder = TreeBuilder;
exports.XMLParser = XMLParser;
exports.elementDefault = elementDefault;
exports.elementDom = dom;
exports.elementSnabbdom = snabbdom;
exports.htmlVoidTagMap = htmlVoidTagMap;
exports.treeRender = treerender;
exports.treeStats = statsEvent;

Object.defineProperty(exports, '__esModule', { value: true });

});
